# Copyright (c) 2024 iiPython

# Modules
import re
import math
from pathlib import Path
from typing import Literal

from mutagen.mp3 import MP3
from mutagen.flac import FLAC
from mutagen.id3._frames import USLT, SYLT

# Initialization
CLASS_MAPPING = {
    ".mp3": MP3, ".flac": FLAC
}
TAG_MAPPING = {
    "TITLE": "TIT2",
    "ALBUM": "TALB",
    "ARTIST": "TPE1",
    "ALBUMARTIST": "TPE2"
}
SYNCED_EXPR = re.compile(r"\[(\d{2}:\d{2}.\d{2})\](.*)")

# Exceptions
class UnsupportedSuffix(ValueError):
    pass

# Audio handler
class AudioFile():
    def __init__(self, path: Path) -> None:
        self.path = path
        if path.suffix not in CLASS_MAPPING:
            raise UnsupportedSuffix(f"Unsupported file extension: '{path.suffix}'!")

        self.file = CLASS_MAPPING[path.suffix](path)
        self.length = self.file.info.length

    def get_tag(self, tag: str) -> str | None:
        if isinstance(self.file, MP3):
            tag = TAG_MAPPING.get(tag, tag)

        if tag in self.file:
            return self.file[tag][0] if isinstance(self.file[tag], list) else self.file[tag]

    def set_tag(self, tag: str, value: str) -> None:
        if isinstance(self.file, MP3):
            tag = TAG_MAPPING.get(tag, tag)

        self.file[tag] = value
        self.file.save()

    def get_lyrics(self, language: str | None = None) -> str | None:
        if isinstance(self.file, FLAC):
            return self.get_tag("LYRICS")

        lyrics = None
        if language is not None:
            lyrics = self.get_tag(f"USLT::{language}") or self.get_tag(f"SYLT::{language}")

        else:

            # Cycle until we find ANY lyrics since we didn't specify
            # a language to look for
            lyrics = [
                value for tag, value in self.file.items()
                if tag[:4] in ["USLT", "SYLT"]
            ]
            lyrics = lyrics[0] if lyrics else None
    
        if isinstance(lyrics, SYLT):
            lyrics = "\n".join([

                # This is absolute garbage and I'll replace it eventually
                f"[{str(math.floor(time / 60000)).zfill(2)}:{str(time / 1000).split('.')[0].zfill(2)}.{str(time / 1000).split('.')[1].zfill(2)}] {text}"
                for (text, time) in lyrics.text  # type: ignore
            ])

        return str(lyrics) if lyrics else None

    def set_lyrics(self, state: Literal["synced"] | Literal["unsynced"], lyrics: str | list, language: str = "XXX") -> None:
        if isinstance(self.file, FLAC):
            if isinstance(lyrics, list):
                raise ValueError

            return self.set_tag("LYRICS", lyrics)

        if state == "synced" and isinstance(lyrics, str):
            new_lyrics = []
            for line in lyrics.split("\n"):
                if not line.strip():
                    continue

                time, text = re.findall(SYNCED_EXPR, line)[0]
                new_lyrics.append((text.strip(), int((60000 * int(time[:2])) + (1000 * float(time[3:])))))

            lyrics = new_lyrics

        arguments = {"lang": language, "text": lyrics}
        if state == "synced":
            arguments |= {"format": 2, "type": 1}

        object = {"unsynced": USLT, "synced": SYLT}[state](**arguments)
        self.set_tag(f"{type(object).__name__}::{language}", object)

# Handle lyrics formatting
def format_lyrics(lyrics: str) -> str:
    return "\n".join([
        f"[{time}] {line}" for (time, line) in re.findall(SYNCED_EXPR, lyrics)
    ])



# Adapted from:
# https://github.com/Dr-Blank/lrclibapi/blob/main/lrclib/cryptographic_challenge_solver.py

# Modules
import hashlib
import threading

# Solution class
class Solution:
    def __init__(self) -> None:
        self.nonce: int = 0

    def is_solved(self) -> bool:
        return self.nonce != 0

def is_nonce_valid(prefix: str, nonce: int, target: bytes) -> bool:
    hash_value = hashlib.sha256(f"{prefix}{nonce}".encode()).digest()
    return hash_value < target

def find_nonce(
    prefix: str,
    target: bytes,
    solution: Solution,
    start: int,
    step: int
) -> Solution:
    nonce = start
    while not solution.is_solved():
        if is_nonce_valid(prefix, nonce, target):
            solution.nonce = nonce
            break

        nonce += step

    return solution

def solve(prefix: str, target: str) -> int:
    target_bytes, solution, threads = bytes.fromhex(target), Solution(), []
    for i in range(4):
        threads.append(threading.Thread(
            target = find_nonce,
            args = (prefix, target_bytes, solution, i, 4),
        ))

    [t.start() for t in threads]
    [t.join() for t in threads]
    return solution.nonce



# Copyright (c) 2024 iiPython

# Modules
from typing import Any, List, Dict

import requests

from . import __version__
from .challenge import solve

# API Controller
class LRCLib():
    def __init__(self, api_url: str = "https://lrclib.net/api/") -> None:
        self.session = requests.Session()
        self.api_url = f"{api_url.rstrip('/')}/"

    def request(self, method: str, endpoint: str, headers: dict = {}, **kwargs) -> requests.Response:
        headers = {
            "User-Agent": f"LRCUP v{__version__} (https://github.com/iiPythonx/lrcup)",
            **headers
        }
        return getattr(self.session, method)(
            self.api_url + endpoint,
            headers = headers,
            **kwargs
        )

    def request_with_404(self, *args, **kwargs) -> dict | list | None:
        resp = self.request(*args, **kwargs).json()
        return None if resp.get("statusCode", 200) == 404 else resp

    def search(
        self,
        query: str | None = None,
        track: str | None = None,
        artist: str | None = None,
        album: str | None = None
    ) -> List[Dict[str, Any]]:
        return self.request("get", "search", params = {
            "q": query,
            "track_name": track,
            "artist_name": artist,
            "album_name": album
        }).json()
    
    def get(
        self,
        track: str,
        artist: str,
        album: str,
        duration: int
    ) -> dict | None:
        return self.request_with_404("get", "get", params = {
            "track_name": track,
            "artist_name": artist,
            "album_name": album,
            "duration": duration
        })

    def get_cached(
        self,
        track: str,
        artist: str,
        album: str,
        duration: int
    ) -> dict | None:
        return self.request_with_404("get", "get-cached", params = {
            "track_name": track,
            "artist_name": artist,
            "album_name": album,
            "duration": duration
        })

    def get_by_id(self, id_: int) -> dict | None:
        return self.request_with_404("get", f"get/{id_}")

    def publish(
        self,
        token: str,
        track: str,
        artist: str,
        album: str,
        duration: int,
        plain_lyrics: str = "",
        synced_lyrics: str = ""
    ) -> bool:
        return self.request(
            "post",
            "publish",
            headers = {"X-Publish-Token": token},
            json = {
                "trackName": track,
                "artistName": artist,
                "albumName": album,
                "duration": duration,
                "plainLyrics": plain_lyrics,
                "syncedLyrics": synced_lyrics
            }
        ).status_code == 201

    def request_challenge(self) -> str:
        data = self.request("post", "request-challenge").json()

        def verify_nonce(result, target) -> bool:
            result_len = len(result)
            if result_len != len(target):
                return False

            for i in range(result_len - 1):
                if result[i] > target[i]:
                    return False

                elif result[i] < target[i]:
                    break

            return True

        nonce = solve(data["prefix"], data["target"])
        return f"{data['prefix']}:{nonce}"



# Copyright (c) 2024 iiPython

# Modules
import sys
from pathlib import Path

import click

from . import __version__
from .audio import AudioFile, UnsupportedSuffix, CLASS_MAPPING, format_lyrics
from .controller import LRCLib

# Initialization
GREEN_FMT = "\033[32m{}\033[0m"

def t(text: str) -> str:
    return f"{text}{' ' * (13 - len(text))}: "

lrclib = LRCLib()

# Input helpers
def update_previous(value: str, prompt: str) -> None:
    print(f"\033[1F\033[2K{prompt}{value}")

def custom_input(prompt: str, value_format: str | None = None) -> str:
    value = input(prompt)
    if value_format is not None:
        update_previous(value_format.format(value), prompt)

    return value

# Handle CLI
@click.group()
def lrcup() -> None:
    """Basic API wrapper for LRCLIB with a CLI included.

    To get started with the module, visit: https://github.com/iiPythonx/lrcup"""
    pass

@lrcup.command(help = "Upload lyrics to LRCLIB from local file")
@click.argument("file", type = click.Path(exists = True, dir_okay = False, path_type = Path))
def upload(file: Path) -> None:
    def process_lyrics(lyrics: str) -> tuple[str, str | None]:
        if all([line.startswith("[") for line in lyrics.split("\n") if line.strip()]):
            print(t("LRC Status"), "\033[32msynced\033[0m", sep = "")
            return "\n".join([line.split("]")[1].lstrip() for line in lyrics.split("\n") if line.strip()]), format_lyrics(lyrics)

        print(t("LRC Status"), "\033[31munsynced\033[0m", sep = "")
        return lyrics, None

    metadata = None

    # Load lyrics format
    if file.suffix in [".txt", ".lrc"]:
        plain_lyrics, synced_lyrics = process_lyrics(file.read_text())

    else:
        try:
            metadata = AudioFile(file)
            lyrics = metadata.get_lyrics()
            if not lyrics:
                return print(t("LRC Status"), "\033[31mmissing\033[0m", sep = "")

            plain_lyrics, synced_lyrics = process_lyrics(lyrics)

        except UnsupportedSuffix as e:
            return click.secho(e, fg = "red")

    # Ask every question known to man
    payload = {}
    for field, readable in [("TITLE", "Track title"), ("ARTIST", "Artist"), ("ALBUM", ("Album", "({TITLE}) "))]:
        addition = ""
        if isinstance(readable, tuple):
            readable, addition = readable

        payload[field] = metadata and metadata.get_tag(field)
        if not payload[field]:  # Catch empty fields as well
            payload[field] = custom_input(t(readable) + addition.format(**payload), GREEN_FMT)
            if field == "ALBUM":
                payload[field] = payload[field] or payload["TITLE"]
                update_previous(GREEN_FMT.format(payload[field]), t(readable))

        else:
            print(t(readable), GREEN_FMT.format(payload[field]), sep = "")

    # Take care of duration
    if metadata is None:
        duration = custom_input(t("Duration") + "(M:S or S) ", GREEN_FMT)
        if ":" in duration:
            duration = duration.split(":")
            duration = (int(duration[0]) * 60) + int(duration[1])

        else:
            duration = int(duration)

        update_previous(GREEN_FMT.format(f"{duration} second(s)"), t("Duration"))

    else:
        duration = metadata.length

    # Confirmation
    if "-y" not in sys.argv:
        if input("\nConfirm upload (y/N)? ") not in ["y", "yes"]:
            return

    # Upload to LRCLIB
    success = lrclib.publish(
        lrclib.request_challenge(),
        payload["TITLE"],
        payload["ARTIST"],
        payload["ALBUM"],
        duration,
        plain_lyrics or "",
        synced_lyrics or ""
    )
    if not success:
        return click.secho("\nFailed to upload to LRCLIB.", fg = "red")

    click.secho("\nUploaded to LRCLIB successfully.", fg = "green")

@lrcup.command(help = "Embed an LRC file into an audio file")
@click.argument("lrc")
@click.argument("destination")
def embed(lrc: str, destination: Path) -> None:
    lyrics = Path(lrc).read_text()
    AudioFile(destination).set_lyrics("synced" if "[" in lyrics else "unsynced", lyrics)

@lrcup.command(help = "Search for specific lyrics by query")
@click.argument("query", nargs = -1, required = True)
def search(query: str) -> None:
    results = []
    for i in lrclib.search(" ".join(query)):
        if not (i["plainLyrics"] or i["syncedLyrics"]):
            continue  # Ignore instrumentals

        results.append(i)

    for i, r in enumerate(results):
        click.echo(f"{i + 1}) {r['artistName'].replace(';', ', ')} - {r['trackName']}")

    if not results:
        return click.secho("No search results found.", fg = "red")

    click.echo()
    try:
        download_id = int(input("ID to Download > "))
        if download_id < 1 or download_id > len(results):
            raise ValueError

        result = results[download_id - 1]
        filename = f"{result['trackName']}.lrc"
        with Path(filename).open("w+") as fh:
            fh.write(result["syncedLyrics"] or result["plainLyrics"])
            click.echo(f"Lyrics written to '{filename}'.")

    except ValueError:
        click.secho("Invalid lyric result ID.", fg = "red")

    except KeyboardInterrupt:
        pass

@lrcup.command(help = "Display version information")
def version() -> None:
    click.echo(f"LRCUP v{__version__} (https://github.com/iiPythonx/lrcup)")

@lrcup.command(help = "Automatically search and embed lyrics for a folder")
@click.argument("target")
@click.option("--force", is_flag = True, show_default = True, default = False, help = "Force searching for lyrics")
def autoembed(target: Path, force: bool) -> None:
    target = Path(target)
    if not target.is_dir():
        return click.secho("Specified target is not a folder.", fg = "red")

    for file in target.rglob("*"):
        if not (file.is_file() and file.suffix in CLASS_MAPPING):
            continue

        try:
            data = AudioFile(file)
            artist, album, title = data.get_tag("ALBUMARTIST") or data.get_tag("ARTIST"), \
                data.get_tag("ALBUM"), data.get_tag("TITLE")

            # I would use all() here but Ruff won't stop complaining
            if not (artist and album and title):
                click.secho(f"[/] Skipping {file} due to missing tags.", fg = "yellow")
                continue

            if data.get_lyrics() and not force:
                click.secho(f"[/] Skipping {title} on {album} by {artist}", fg = "yellow")
                continue

            # Perform lyrics search
            results = lrclib.get(title, artist, album, round(data.length))
            if not results:
                click.secho(f"[-] No results found for {title} on {album} by {artist}", fg = "red")
                continue

            lyrics = (results.get("syncedLyrics") or results.get("plainLyrics")) or ""
            if not lyrics.strip():
                click.secho(f"[-] No results found for {title} on {album} by {artist}", fg = "red")
                continue

            data.set_lyrics("synced" if "[" in lyrics else "unsynced", lyrics)
            click.secho(f"[+] Fetched lyrics for {title} on {album} by {artist}", fg = "green")

        except Exception:
            click.secho(f"[-] Failed to read tags from file '{file}'", fg = "red")

if __name__ == "__main__":
    lrcup()
